#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) 2015-2019 Chintalagiri Shashank
#
# This file is part of tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Reusable minor helpers for www (:mod:`tendril.utils.www.helpers`)
=================================================================

TODO Some introduction

"""


from tendril.config import NETWORK_PROXY_TYPE
from tendril.config import NETWORK_PROXY_IP
from tendril.config import NETWORK_PROXY_PORT
from tendril.config import NETWORK_PROXY_USER
from tendril.config import NETWORK_PROXY_PASS

from tendril.utils import log
logger = log.get_logger(__name__, log.WARNING)


def get_http_proxy_url():
    """
    Constructs the proxy URL for HTTP proxies from relevant
    :mod:`tendril.config` Config options, and returns the URL string
    in the form:

        ``http://[NP_USER:NP_PASS@]NP_IP[:NP_PORT]``

    where NP_xxx is obtained from the :mod:`tendril.config` ConfigOption
    NETWORK_PROXY_xxx.
    """
    if NETWORK_PROXY_USER is None:
        proxyurl_http = 'http://' + NETWORK_PROXY_IP
    else:
        proxyurl_http = 'http://{0}:{1}@{2}'.format(NETWORK_PROXY_USER,
                                                    NETWORK_PROXY_PASS,
                                                    NETWORK_PROXY_IP)
    if NETWORK_PROXY_PORT:
        proxyurl_http += ':' + NETWORK_PROXY_PORT
    return proxyurl_http


def _get_proxy_dict():
    """
    Construct a dict containing the proxy settings in a format compatible
    with the :class:`requests.Session`. This function is used to construct
    the :data:`_proxy_dict`.

    """
    if NETWORK_PROXY_TYPE == 'http':
        proxyurl = get_http_proxy_url()
        return {'http': proxyurl,
                'https': proxyurl}
    else:
        return None


#: A dict containing the proxy settings in a format compatible
#: with the :class:`requests.Session`.
proxy_dict = _get_proxy_dict()


def strencode(string):
    """
    This function converts unicode strings to ASCII, using python's
    :func:`str.encode`, replacing any unicode characters present in the
    string. Unicode characters which Tendril expects to see in web content
    related to it are specifically replaced first with ASCII characters
    or character sequences which reasonably reproduce the original meanings.

    :param string: unicode string to be encoded.
    :return: ASCII version of the string.

    .. warning::
        This function is marked for deprecation by the general (but gradual)
        move towards ``unicode`` across tendril.

    """
    nstring = ''
    for char in string:
        if char == u'\u00b5':
            char = 'u'
        if char == u'\u00B1':
            char = '+/-'
        nstring += char
    return nstring.encode('ascii', 'replace')





from tendril.utils.config import ConfigOption
from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)

depends = ['tendril.config.core']


config_elements_network_caching = [
    ConfigOption(
        'ENABLE_REDIRECT_CACHING',
        "False",
        "Whether or not to cache 301 and 302 redirects. "
        "This probably works just fine, but treat it as a largely "
        "deprecated feature."
    ),
    ConfigOption(
        'MAX_AGE_DEFAULT',
        '600000',
        'Default max_age for data originating from www.'
    ),
]


config_elements_proxy = [
    ConfigOption(
        'NETWORK_PROXY_TYPE',
        "None",
        "The type of proxy to use. 'http' for squid/http, 'None' for none."
        "No other proxy types presently supported."
    ),
    ConfigOption(
        'NETWORK_PROXY_IP',
        "None",
        "The proxy server IP address."
    ),
    ConfigOption(
        'NETWORK_PROXY_PORT',
        "3128",
        "The proxy server port."
    ),
    ConfigOption(
        'NETWORK_PROXY_USER',
        "None",
        "The username to authenticate with the proxy server."
    ),
    ConfigOption(
        'NETWORK_PROXY_PASS',
        "None",
        "The password to authenticate with the proxy server."
    ),
]


config_elements_ssl = [
    ConfigOption(
        'CA_BUNDLE',
        "None",
        "Path to a custom CA certificate bundle to use. This is "
        "presently only used by the httpx backend"
    ),
    ConfigOption(
        'SSL_NOVERIFY_HOSTS',
        "[]",
        "List of domains in the form 'https://<domain>' for which "
        "SSL verification is to be skipped. This is intended for "
        "development purposes only."
    ),
]


def load(manager):
    logger.debug("Loading {0}".format(__name__))
    manager.load_elements(config_elements_network_caching,
                          doc="Network Caching Behavior Configuration")
    manager.load_elements(config_elements_proxy,
                          doc="Network Proxy Configuration")
    manager.load_elements(config_elements_ssl,
                          doc="SSL Client Configuration")



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) 2015-2019 Chintalagiri Shashank
#
# This file is part of tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
suds based www SOAP backend (:mod:`tendril.utils.www.soap`)
===========================================================

TODO Some introduction

"""


import os
import six
import time
import logging
from hashlib import md5
from suds.client import Client
from suds.transport.http import HttpAuthenticated
from suds.transport.http import HttpTransport

try:
    import cPickle as pickle
except ImportError:
    import pickle

from .helpers import proxy_dict
from .caching import CacheBase

from tendril.config import INSTANCE_CACHE
from tendril.config import MAX_AGE_DEFAULT

from tendril.utils import log
logger = log.get_logger(__name__, log.WARNING)

logging.getLogger('suds.xsd.query').setLevel(logging.INFO)
logging.getLogger('suds.xsd.sxbasic').setLevel(logging.INFO)
logging.getLogger('suds.xsd.schema').setLevel(logging.INFO)
logging.getLogger('suds.xsd.sxbase').setLevel(logging.INFO)
logging.getLogger('suds.metrics').setLevel(logging.INFO)
logging.getLogger('suds.wsdl').setLevel(logging.INFO)
logging.getLogger('suds.client').setLevel(logging.INFO)
logging.getLogger('suds.resolver').setLevel(logging.INFO)
logging.getLogger('suds.umx.typed').setLevel(logging.INFO)
logging.getLogger('suds.mx.literal').setLevel(logging.INFO)
logging.getLogger('suds.mx.core').setLevel(logging.INFO)
logging.getLogger('suds.transport.http').setLevel(logging.INFO)

SOAP_CACHE = os.path.join(INSTANCE_CACHE, 'soapcache')


class ThrottledTransport(HttpAuthenticated):
    def __init__(self, **kwargs):
        """
        Provides a throttled HTTP transport for respecting rate limits
        on rate-restricted SOAP APIs using :mod:`suds`.

        This class is a :class:`suds.transport.Transport` subclass
        based on the default ``HttpAuthenticated`` transport.

        :param minimum_spacing: Minimum number of seconds between requests.
                                Default 0.
        :type minimum_spacing: int

        .. todo::
            Use redis or so to coordinate between threads to allow a
            maximum requests per hour/day limit.

        """
        self._minumum_spacing = kwargs.pop('minimum_spacing', 0)
        self._last_called = int(time.time())
        HttpTransport.__init__(self, **kwargs)

    def send(self, request):
        """
        Send a request and return the response. If the minimum number of
        seconds between requests have not yet elapsed, then the function
        sleeps for the remaining period and then passes the request along.
        """
        now = int(time.time())
        logger.debug('Getting SOAP response')
        tsincelast = now - self._last_called
        if tsincelast < self._minumum_spacing:
            tleft = self._minumum_spacing - tsincelast
            logger.info("Throttling SOAP client for {0}".format(tleft))
            time.sleep(tleft)
        self._last_called = now
        return HttpAuthenticated.send(self, request)


class CachedTransport(CacheBase, HttpAuthenticated):
    def __init__(self, **kwargs):
        """
        Provides a cached HTTP transport with request-based caching for
        SOAP APIs using :mod:`suds`.

        This is a subclass of :class:`CacheBase` and the default
        ``HttpAuthenticated`` transport.

        :param cache_dir: folder where the cache is located.
        :param max_age: the maximum age in seconds after which a response
                        is considered stale.

        """
        cache_dir = kwargs.pop('cache_dir')
        self._max_age = kwargs.pop('max_age', MAX_AGE_DEFAULT)
        CacheBase.__init__(self, cache_dir=cache_dir)

    def _get_filepath(self, request):
        """
        Return a filename constructed from the md5 hash of a
        combination of the request URL and message content
        (encoded as ``utf-8`` if necessary).

        :param request: the request object for which a cache filename
                        is needed.
        :return: name of the cache file.

        """
        keystring = request.url + request.message
        if six.PY3 or (six.PY2 and isinstance(keystring, unicode)):  # noqa
            filepath = md5(keystring.encode('utf-8')).hexdigest()
        else:
            filepath = md5(keystring).hexdigest()
        return filepath

    def _get_fresh_content(self, request):
        """
        Retrieve a fresh copy of the resource from the source.

        :param request: the request object for which the response
                        is needed.
        :return: the response to the request

        """
        response = HttpAuthenticated.send(self, request)
        return response

    @staticmethod
    def _serialize(response):
        """
        Serializes the suds response object using :mod:`cPickle`.

        If the response has an error status (anything other than
        200), raises ``ValueError``. This is used to avoid caching
        errored responses.

        """
        if response.code != 200:
            logger.debug("Bad Status {0}".format(response.code))
            raise ValueError
        return pickle.dumps(response)

    @staticmethod
    def _deserialize(filecontent):
        """
        De-serializes the cache content into a suds response object
        using :mod:`cPickle`.

        """
        return pickle.loads(filecontent)

    def send(self, request):
        """
        Send a request and return the response. If a fresh response to the
        request is available in the cache, that is returned instead. If it
        isn't, a fresh response is obtained, cached, and returned.

        """
        response = self._accessor(self._max_age, False, request)
        return response


class CachedThrottledTransport(ThrottledTransport, CachedTransport):
    def __init__(self, **kwargs):
        """
        A cached HTTP transport with both throttling and request-based
        caching for SOAP APIs using :mod:`suds`.

        This is a subclass of :class:`CachedTransport` and
        :class:`ThrottledTransport`.

        Keyword arguments not handled here are passed on via
        :class:`ThrottledTransport` to :class:`HttpTransport`.

        :param cache_dir: folder where the cache is located.
        :param max_age: the maximum age in seconds after which a response
                        is considered stale.
        :param minimum_spacing: Minimum number of seconds between requests.
                                Default 0.
        """
        cache_dir = kwargs.pop('cache_dir')
        max_age = kwargs.pop('max_age', MAX_AGE_DEFAULT)
        CachedTransport.__init__(self, cache_dir=cache_dir, max_age=max_age)
        ThrottledTransport.__init__(self, **kwargs)

    def _get_fresh_content(self, request):
        """
        Retrieve a fresh copy of the resource from the source via
        :func:`ThrottledTransport.send`.

        :param request: the request object for which the response
                        is needed.
        :return: the response to the request

        """
        response = ThrottledTransport.send(self, request)
        return response

    def send(self, request):
        """
        Send a request and return the response, using
        :func:`CachedTransport.send`.

        """
        return CachedTransport.send(self, request)


def get_soap_client(wsdl, cache_requests=True,
                    max_age=MAX_AGE_DEFAULT, minimum_spacing=0):
    """
    Creates and returns a suds/SOAP client instance bound to the
    provided ``WSDL``. If ``cache_requests`` is True, then the
    client is configured to use a :class:`CachedThrottledTransport`.
    The transport is constructed to use :data:`SOAP_CACHE` as the
    cache folder, along with the ``max_age`` and ``minimum_spacing``
    parameters if provided.

    If ``cache_requests`` is ``False``, the client uses the default
    :class:`suds.transport.http.HttpAuthenticated` transport.
    """
    if cache_requests is True:
        if proxy_dict is None:
            soap_transport = CachedThrottledTransport(
                cache_dir=SOAP_CACHE, max_age=max_age,
                minimum_spacing=minimum_spacing,
            )
        else:
            soap_transport = CachedThrottledTransport(
                cache_dir=SOAP_CACHE, max_age=max_age,
                minimum_spacing=minimum_spacing,
                proxy=proxy_dict,
            )
    else:
        if proxy_dict is None:
            soap_transport = HttpAuthenticated()
        else:
            soap_transport = HttpAuthenticated(proxy=proxy_dict)
    return Client(wsdl, transport=soap_transport)



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) 2015-2019 Chintalagiri Shashank
#
# This file is part of tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
requests based www backend (:mod:`tendril.utils.www.bare`)
==========================================================

TODO Some introduction

"""

import os
import logging
import requests
from cachecontrol import CacheControlAdapter
from cachecontrol.caches import FileCache
from cachecontrol.heuristics import ExpiresAfter
from bs4 import BeautifulSoup

from tendril.config import INSTANCE_CACHE
from tendril.config import MAX_AGE_DEFAULT

from .helpers import proxy_dict

from tendril.utils import log

logger = log.get_logger(__name__, log.WARNING)

logging.getLogger('cachecontrol.controller').setLevel(logging.INFO)
logging.getLogger('requests.packages.urllib3.connectionpool').\
    setLevel(logging.WARNING)

REQUESTS_CACHE = os.path.join(INSTANCE_CACHE, 'requestscache')


#: The module's :class:`cachecontrol.caches.FileCache` instance which
#: should be used whenever cached :mod:`requests` responses are desired.
#: The cache is stored in the directory defined by
#: :data:`tendril.config.REQUESTS_CACHE`.
#: This cache uses very weak permissions. These should probably be
#: fine tuned.
requests_cache = FileCache(REQUESTS_CACHE, filemode=0o666, dirmode=0o777)


def _get_requests_cache_adapter(heuristic):
    """
    Given a heuristic, constructs and returns a
    :class:`cachecontrol.CacheControlAdapter` attached to the instance's
    :data:`requests_cache`.

    """
    return CacheControlAdapter(
        cache=requests_cache,
        heuristic=heuristic,
        cache_etags=False
    )


def get_session(target='http://', heuristic=None):
    """
    Gets a pre-configured :mod:`requests` session.

    This function configures the following behavior into the session :

    - Proxy settings are added to the session.
    - It is configured to use the instance's :data:`requests_cache`.
    - Permanent redirect caching is handled by :mod:`CacheControl`.
    - Temporary redirect caching is not supported.

    Each module / class instance which uses this should subsequently
    maintain it's own session with whatever modifications it requires
    within a scope which makes sense for the use case (and probably close
    it when it's done).

    The session returned from here uses the instance's REQUESTS_CACHE with
    a single - though configurable - heuristic. If additional caches or
    heuristics need to be added, it's the caller's problem to set them up.

    .. note::
        The caching here seems to be pretty bad, particularly for digikey
        passive component search. I don't know why.

    :param target: Defaults to ``'http://'``. string containing a prefix
                   for the targets that should be cached. Use this to setup
                   site-specific heuristics.
    :param heuristic: The heuristic to use for the cache adapter.
    :type heuristic: :class:`cachecontrol.heuristics.BaseHeuristic`
    :rtype: :class:`requests.Session`

    """

    s = requests.session()
    if proxy_dict is not None:
        s.proxies.update(proxy_dict)
    if heuristic is None:
        heuristic = ExpiresAfter(seconds=MAX_AGE_DEFAULT)
    s.mount(target, _get_requests_cache_adapter(heuristic))
    return s


def get_soup_requests(url, session=None):
    """
    Gets a :mod:`bs4` parsed soup for the ``url`` specified by the parameter.
    The :mod:`lxml` parser is used.

    If a ``session`` (previously created from :func:`get_session`) is
    provided, this session is used and left open. If it is not, a new session
    is created for the request and closed before the soup is returned.

    Using a caller-defined session allows re-use of a single session across
    multiple requests, therefore taking advantage of HTTP keep-alive to
    speed things up. It also provides a way for the caller to modify the
    cache heuristic, if needed.

    Any exceptions encountered will be raised, and are left for the caller
    to handle. The assumption is that a HTTP or URL error is going to make
    the soup unusable anyway.

    """
    if session is None:
        session = get_session()
        _close_after = True
    else:
        _close_after = False

    r = session.get(url)
    r.raise_for_status()
    soup = BeautifulSoup(r.content, 'lxml', from_encoding=r.encoding)

    if _close_after is True:
        session.close()
    return soup



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) 2015-2019 Chintalagiri Shashank
#
# This file is part of tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Redirect Caching Infrastructure (:mod:`tendril.utils.www.redirectcaching`)
==========================================================================

TODO Some introduction

"""


import os
import atexit

try:
    import cPickle as pickle
except ImportError:
    import pickle

from six.moves.urllib.request import HTTPRedirectHandler

from tendril.config import INSTANCE_CACHE
from tendril.config import ENABLE_REDIRECT_CACHING

from tendril.utils import log
logger = log.get_logger(__name__, log.WARNING)


REDIR_CACHE_FILE = os.path.join(INSTANCE_CACHE, 'redirects.p')

if ENABLE_REDIRECT_CACHING:
    logger.info("Initializing Redirect Caching")
    try:
        with open(REDIR_CACHE_FILE, "rb") as rdcf:
            redirect_cache = pickle.load(rdcf)
        logger.info('Loaded Redirect Cache from file')
    except EOFError:
        os.remove(REDIR_CACHE_FILE)
        redirect_cache = {}
        logger.warning('Discarded Corrupt Redirect Cache')
    except (IOError, FileNotFoundError):
        redirect_cache = {}
        logger.info('Created new Redirect Cache')
else:
    redirect_cache = None


def dump_redirect_cache():
    """
    Called during python interpreter shutdown, this function dumps the
    redirect cache to the file system.
    """
    if DUMP_REDIR_CACHE_ON_EXIT:
        with open(REDIR_CACHE_FILE, 'wb') as f:
            pickle.dump(redirect_cache, f, protocol=2)
        logger.info('Dumping Redirect Cache to file')


if ENABLE_REDIRECT_CACHING:
    DUMP_REDIR_CACHE_ON_EXIT = True
else:
    DUMP_REDIR_CACHE_ON_EXIT = False

if ENABLE_REDIRECT_CACHING is True:
    atexit.register(dump_redirect_cache)


class CachingRedirectHandler(HTTPRedirectHandler):
    """
    This handler modifies the behavior of
    :class:`urllib2.HTTPRedirectHandler`, resulting in a HTTP ``301`` or
    ``302`` status to be included in the ``result``.

    When this handler is attached to a ``urllib2`` opener, if the opening of
    the URL resulted in a redirect via HTTP ``301`` or ``302``, this is
    reported along with the result. This information can be used by the opener
    to maintain a redirect cache.
    """
    def __init__(self):
        pass

    def http_error_301(self, req, fp, code, msg, headers):
        """
        Wraps the :func:`urllib2.HTTPRedirectHandler.http_error_301` handler,
        setting the ``result.status`` to ``301`` in case a http ``301`` error
        is encountered.
        """
        result = HTTPRedirectHandler.http_error_301(
            self, req, fp, code, msg, headers)
        result.status = code
        return result

    def http_error_302(self, req, fp, code, msg, headers):
        """
        Wraps the :func:`urllib2.HTTPRedirectHandler.http_error_302` handler,
        setting the ``result.status`` to ``302`` in case a http ``302`` error
        is encountered.
        """
        result = HTTPRedirectHandler.http_error_302(
            self, req, fp, code, msg, headers)
        result.status = code
        return result


def get_actual_url(url):
    # warnings.warn("get_actual_url() is a part of Redirect caching and is "
    #               "deprecated.", DeprecationWarning)
    if not ENABLE_REDIRECT_CACHING:
        return url
    else:
        while url in redirect_cache.keys():
            url = redirect_cache[url]
        return url
